@@ -105,11 +105,13 @@ static const FUNMAP default_funmap[] = {
   { "kill-region", rl_kill_region },
   { "kill-word", rl_kill_word },
   { "menu-complete", rl_menu_complete },
+  { "menu-complete-backward", rl_backward_menu_complete },
   { "next-history", rl_get_next_history },
   { "non-incremental-forward-search-history", rl_noninc_forward_search },
   { "non-incremental-reverse-search-history", rl_noninc_reverse_search },
   { "non-incremental-forward-search-history-again", rl_noninc_forward_search_again },
   { "non-incremental-reverse-search-history-again", rl_noninc_reverse_search_again },
+  { "old-menu-complete", rl_old_menu_complete },
   { "overwrite-mode", rl_overwrite_mode },
 #ifdef __CYGWIN__
   { "paste-from-clipboard", rl_paste_from_clipboard },
@@ -123,6 +125,7 @@ static const FUNMAP default_funmap[] = {
   { "revert-line", rl_revert_line },
   { "self-insert", rl_insert },
   { "set-mark", rl_set_mark },
+  { "skip-csi-sequence", rl_skip_csi_sequence },
   { "start-kbd-macro", rl_start_kbd_macro },
   { "tab-insert", rl_tab_insert },
   { "tilde-expand", rl_tilde_expand },
